function love.load()

-- Finishing rectangle
	math.randomseed( os.time() )
	-- Stage
	stage = {}
	curStage = 0
	
	-- Time between beginning of level
	-- and the ball dropping.
	dropTimer = 4
	
	won = false
	lose = false
	timeRemaining = 34
		
	introStr     = "Welcome to Ninwa's Love Puzzle Game\n\n" ..
	               "     How it works:\n\n"..
	               "          You can click anywhere on the screen to define a \n"..
	               "          new node. Each node will automatically 'attach' \n"..
	               "          to the next nearest node. A ball will drop from the\n"..
	               "          sky and the objective is to guide it into the finish.\n\n"..
	               "          If failure seems inevitable, press 'r' to reset the level.\n\n"..
	               "There's one catch though, each level you only get a handful\n"..
	               "of nodes to use, and you must use the nodes from the previous\n"..
	               "level. You may at anytime revisit and old stage, but you must\n"..
	               "then complete the subsequent stages. Good luck! \n\n"..
	               "How far can you get? :)\n\n-Ninwa\n\n\n(Press 'space' to begin)"
	               
	introduction = true
	game = false
	
	-- Hold all stages finishes boxes
	-- so we can reserve its space when
	-- we go back
	finishingRects = {}
	for i=1, 99 do 
		table.insert(finishingRects, {x = math.random(0,400),
		                              y = math.random(250,400),
									  w = 100, h = 100})
	end
	 
				 				  
end

-- Setup a stage, if no argument is passed
-- it generates the next stage, if s is not nil
-- then we load that stage.
function newStage(s,r)
	if s ~= nil then
		curStage = s
	else
		curStage = curStage + 1
	end
	
	-- Setup love.physics
	world = love.physics.newWorld(500,500)
	world:setGravity(0,50)
	world:setMeter(50)
	ball  = love.physics.newBody(world, 250,  50, 15, 0) 
	shape = love.physics.newCircleShape(ball, 0, 0, 20)  
	
	ball:setMassFromShapes()
	nodes = {}
	
	if curStage > 1 then	
	
		if s ~= nil and r == nil then
			-- Copy all nodes from a specific stage
			for k, v in ipairs(stage[s].nodes) do
				local x = v.x
				local y = v.y
				table.insert(nodes,{["x"] = v.x, ["y"] = v.y})
			end
		else
			-- Copy all nodes from previous stage
			for k, v in ipairs(stage[curStage-1].nodes) do
				local x = v.x
				local y = v.y
				table.insert(nodes,{["x"] = v.x, ["y"] = v.y})
			end
		end
		checkNeighbors()
	else
		bodies = {}
		shapes = {}
		nodes  = {}
	end
	
	nodesRemaining = 4
	
	
	if s ~= nil then
		finishRect = finishingRects[s]
	else
		finishRect = finishingRects[curStage]
	end
	
	             
	timeEnteredFinish = 0 
	inFinish = false
	won = false
	lose = false
	
	dropTimer = love.timer.getTime() + 4
	timeRemaining = love.timer.getTime() + 33
	saveStage(curStage)
end

-- Takes all of our world data and 
-- stores them in a table where n
-- is the index
function saveStage(n)
	stage[n] = {
			    ["world"]  = world,
			    ["ball"]   = ball,
			    ["shape"]  = shape,
			    ["bodies"] = bodies,
			    ["finishRect"] = finishRect,
			    ["nodes"] = nodes
			    
			   }
end

function checkNeighbors()
	-- Connect each node to it's nearest neighbor
	-- and create/destroy the appropriate box2d rects
	for ko,vo in ipairs(nodes) do
		vo.attachedNode = nil
		local bestDistance = nil
		local bestCoords   = nil
		
		if vo.body ~= nil and vo.shape ~= nil then
			vo.body:destroy()
			vo.shape:destroy()
			vo.body = nil
			vo.shape = nil
		end
		
		for ki, vi in ipairs(nodes) do
			local distance = math.sqrt((vi.x-vo.x)^2+(vi.y-vo.y)^2)
			if ki ~= ko and 
			   (bestDistance == nil or distance < bestDistance) then
				bestDistance = distance
				bestCoords = vi
			end
		end 
		
		if bestCoords ~= nil then
			vo.attachedNode = bestCoords
			
			if vo.body == nil then
				local a = math.atan((bestCoords.y-vo.y)/
				                    (bestCoords.x-vo.x))	
				vo.body =    love.physics.newBody(world,vo.x,vo.y,0,0)

				vo.shape = love.physics.newRectangleShape(vo.body,
				                                          (bestCoords.x-vo.x)/2,
				                                          (bestCoords.y-vo.y)/2,
														  bestDistance,
														  5,a)
			end
		end
	end
end

function love.update(dt)
	if introduction then
		return
	end
	
	if game == false then
		game = true
		newStage()
	end
	
	-- Pause all physics until our ball drop timer is complete
	if dropTimer - love.timer.getTime() <= 0 then
	
		-- Make sure we didn't lose!
		if ball:getX() < 500 and ball:getX() > 0 and
		   ball:getY() < 500 and 
		   timeRemaining - love.timer.getTime() > 0  then
		   
		   -- Freeze all movement when we've won, too!
		   if not won then
				-- Allow gravity to act upon our body
				world:update(dt)
			end
		else
			if not won then
				lose = true
			end
		end
	end
	
	-- Determine if our ball has entered our finishing rect
	-- and if it has, set the time we've entered. If we leave
	-- the rect make sure we unset the time entered.
	if ball:getX() > finishRect.x and
	   ball:getY() > finishRect.y and
	   ball:getX() < finishRect.x + finishRect.w and
	   ball:getY() < finishRect.y + finishRect.h then
	
		if not inFinish  then
			inFinish = true
			
			if timeEnteredFinish == 0 then
				timeEnteredFinish = love.timer.getTime()
			end
		end
	else
		inFinish = false
		timeEnteredFinish = 0
	end
	
	if (love.timer.getTime() - timeEnteredFinish) >= 5  and
		inFinish == true and timeEnteredFinish ~= 0 then
		won = true
	end
end

function love.mousepressed(x,y,button)
	if introduction then
		return
	end

	-- Insert a node when the left mouse button is clicked
	if button == 'l' and not won and not lose then
		if nodesRemaining > 0 then
			table.insert(nodes,{["x"] = x,["y"] = y, 
		                    ["body"] = nil, ["shape"] = nil})
			checkNeighbors()
			nodesRemaining = nodesRemaining - 1
		end
	end
end

function love.keypressed(key,unicode)
	if key == " " then
	
		if introduction then
			introduction = false
		end
		
		if won then
			-- Generate new stage
			newStage()
		end
		
		if lose then
			if curStage > 1 then
				newStage(curStage,true)
			else
				newStage(1)
			end
			lose = false
		end
	end
	
	if key == 'left' and lose then
		if curStage > 1 then
			curStage = curStage - 1
			newStage(curStage,true)
		end
	end
	
	if introduction then
		return
	end
	
	if key == "r" and dropTimer - love.timer.getTime() <= 0  and not won and not lose then
		timeRemaining = love.timer.getTime()-1 -- hacks :p
	end
end

function love.draw()

	if introduction then
		love.graphics.print(introStr,20,20)
		return
	end

	love.graphics.setColor(255,255,255)
	displayStr = "Love Physics Puzzle\nby Ninwa"
	love.graphics.print(displayStr,20,20)
	
	-- Draw time remaining
	if not won and not lose then
		trStr = math.floor(timeRemaining - love.timer.getTime()) .. 
				" seconds remaining..."
		love.graphics.print(trStr, 340, 462)
	end
		
	-- Draw stage and nodes remaining
	statusStr = "Stage: " .. curStage .. "\nNodes Remaining: " .. 
	            nodesRemaining
	love.graphics.print(statusStr, 20, 460)
	
	-- Draw nodes and lines
	for i, v in ipairs(nodes) do
		love.graphics.circle('line',v.x,v.y,5,1024)
		if v.attachedNode ~= nil then
			love.graphics.line(v.x,v.y,
			                   v.attachedNode.x,v.attachedNode.y)
		end
	end
	
	-- Draw our ball
	love.graphics.circle("line", ball:getX(), ball:getY(), 
		                  shape:getRadius(), 20)
	-- Draw finishing goal
	love.graphics.setColor(0,200,100)
	love.graphics.rectangle('line', finishRect.x, finishRect.y, 
	                        finishRect.w, finishRect.h)
	                        
	-- Draw ball drop timer
	local timerStr = math.floor(dropTimer - love.timer.getTime())
	if dropTimer - love.timer.getTime() > 0 then
		love.graphics.print(timerStr.."...",240,250)
	end 
	
	-- Draw time in finish if applicable
	if inFinish and not won then
		local finishMsg = "You've made it, you need to stay there:\n" ..
		                   math.floor(5 - (love.timer.getTime() - timeEnteredFinish))+1 ..
		                   " more seconds!"
		love.graphics.print(finishMsg, 200,20)
	end
	
	if won then
		local wonMsg = "You've won!\n\nPress 'space' to continue"
		love.graphics.print(wonMsg, 170,150)
	end
	
	if lose then
		local loseMsg = "Aw, didn't make it!\n\nPress 'space' to try again!\n\n"
		love.graphics.print(loseMsg, 170,150)
		
		if curStage > 1 then
			local backMsg = "\n... or press the left arrow key to go back a stage!"
			love.graphics.print(backMsg,170,190)
		end
	end
end
